import pytest
from poetry_analysis.rhyme_detection import remove_syllable_onset


@pytest.mark.parametrize("syllable, expected", [
    (["H", "AH2", "L", "OUH0"], ["AH2", "L", "OUH0"]),
    (["V",  "AE1",  "RD",  "RNX0"], ["AE1", "RD", "RNX0"]),
    (["D", "UU1"], ["UU1"] ),
])
def test_onset_is_removed_for_syllables(syllable, expected):
    result = remove_syllable_onset(syllable)
    assert result == expected


@pytest.mark.parametrize("syllable", [
    "hello world", "H A L L O", "H EI", ["B", "P", "K", "L"]
])
def test_returns_none_when_no_nucleus_is_found(syllable):
    result = remove_syllable_onset(syllable)
    assert result is None



# %%
# rhyme_detection main module!

import re
import string
import logging
import json
from pathlib import Path

from dataclasses import dataclass
from convert_pa import convert_nofabet

from poetry_analysis import utils


@dataclass
class Poem:
    _id: str
    text: str
    stanzas: list


@dataclass
class Verse:
    _id: str
    text: str
    tokens: list
    transcription: str
    syllables: list
    rhyme_tag: str


def is_stressed(syllable: str | list) -> bool:
    """Check if a syllable is stressed by searching for stress markers.

    Stress markers:
        0 - Vowel/syllable nucleus without stress
        1 - Primary stress with toneme 1
        2 - Primary stress with toneme 2
        3 - Secondary stress
    """
    if isinstance(syllable, list):
        syllable = " ".join(syllable)
    result = re.search(r"[123]", syllable)
    return True if result else False


def strip_stress(phoneme: str) -> str:
    """Strip the stress marker from a phoneme."""
    return phoneme.strip("0123")


def is_nucleus(phoneme: str) -> bool:
    """Check if a phoneme is a vowel."""
    syll_nuclei = convert_nofabet.PHONES_NOFABET.get("nuclei")
    return True if strip_stress(phoneme) in syll_nuclei else False


def find_last_stressed_syllable(syllables: list) -> list:
    for idx, syll in reversed(list(enumerate(syllables))):
        if is_stressed(syll):
            # flatten the rhyming syllable sequence
            stressed = [s for rhyme in syllables[idx:] for s in rhyme]
            return stressed


def find_syllable_rhyme(syllables: list) -> list:
    """Identify the rhyming part of a verse.

    Args:
        syllables: nested list of lists of phonemes
    """
    stressed = find_last_stressed_syllable(syllables)
    return remove_syllable_onset(stressed)


def remove_syllable_onset(syllable: list) -> list:
    """Split a syllable nucleus and coda from the onset to find the rhyming part of the syllable."""
    for idx, phone in enumerate(syllable):
        if is_nucleus(phone):
            return syllable[idx:]
    logging.debug("No nucleus found in %s", syllable)


"""
TODO: Fiks feil med NoneType

Traceback (most recent call last):
  File "/home/ingeridd/prosjekter/NORN-poems/src/poetry_analysis/rhyme_detection.py", line 264, in <module>
    main(args.jsonfile)
  File "/home/ingeridd/prosjekter/NORN-poems/src/poetry_analysis/rhyme_detection.py", line 229, in main
    tagged = tag_rhyming_verses(stanza)
             ^^^^^^^^^^^^^^^^^^^^^^^^^^
  File "/home/ingeridd/prosjekter/NORN-poems/src/poetry_analysis/rhyme_detection.py", line 144, in tag_rhyming_verses
    rhyme_score = score_rhyme(previous_syll, current_syllables)
                  ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  File "/home/ingeridd/prosjekter/NORN-poems/src/poetry_analysis/rhyme_detection.py", line 102, in score_rhyme
    if do_syll_seqs_rhyme(last_syll1, last_syll2):
       ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  File "/home/ingeridd/prosjekter/NORN-poems/src/poetry_analysis/rhyme_detection.py", line 87, in do_syll_seqs_rhyme
    if all(strip_stress(s1) == strip_stress(s2) for s1, s2 in zip(syll1, syll2)):
                                                              ^^^^^^^^^^^^^^^^^
TypeError: 'NoneType' object is not iterable
"""


def do_syll_seqs_rhyme(syll1: list, syll2: list):
    """Check  if each syllable in two syllable sequences are identical, apart from the stress marker."""
    if all(strip_stress(s1) == strip_stress(s2) for s1, s2 in zip(syll1, syll2)):
        return True
    return False


def score_rhyme(syllable1: list, syllable2: list) -> int:
    """Check if two syllable sequences rhyme, and return a rhyming score.

    If the onset, nucleus and coda of two different syllable sequences are the same,
        the score is 0.5. (Nødrim, e.g. "fryd" and "fryd")
    If the rhyming parts (without onsets) of two syllable sequences have the same phonemes, the score is 1.
    If they don't match at all, the score is 0.
    """
    last_syll1 = find_last_stressed_syllable(syllable1)
    last_syll2 = find_last_stressed_syllable(syllable2)
    try:
        is_rhyming = do_syll_seqs_rhyme(last_syll1, last_syll2)
    except TypeError:
        logging.error("Error in syllable comparison: %s and %s", last_syll1, last_syll2)
        return 0
    if is_rhyming:
        logging.debug("NØDRIM: %s and %s", last_syll1, last_syll2)
        return 0.5

    rhyme1 = find_syllable_rhyme(syllable1)
    rhyme2 = find_syllable_rhyme(syllable2)
    try:
        is_rhyming = do_syll_seqs_rhyme(rhyme1, rhyme2)
    except TypeError:
        logging.error("Error in syllable comparison: %s and %s", rhyme1, rhyme2)
        return 0
    if is_rhyming:
        logging.debug("Rhyme: %s and %s", rhyme1, rhyme2)
        rhyme_score = 1
    else:
        # logging.debug("No rhyme: %s and %s", rhyme1, rhyme2)
        rhyme_score = 0
    return rhyme_score


def tag_rhyming_verses(transcribed_verses: list) -> list:
    """Annotate end rhyme patterns in a poem stanza.

    Args:
        transcribed_verses: list of phonemically transcribed and syllabified verse lines
    Return:
        list of annotated verses with transcriptions and rhyme tags
    """
    alphabet = iter(string.ascii_letters)

    processed = []  # needs to be a list!
    # iteratively compare previous lines with current line
    for idx, current_syllables in enumerate(transcribed_verses):
        if not current_syllables:  # skip empty lines
            continue
        does_rhyme = False
        for previous_verse in reversed(processed):
            previous_syll = previous_verse.get("syllables")
            rhyme_score = score_rhyme(previous_syll, current_syllables)
            if rhyme_score > 0:
                rhyme_tag = previous_verse.get("rhyme_tag")
                does_rhyme = True
                break

        if not does_rhyme:
            rhyme_score = 0
            try:
                rhyme_tag = next(alphabet)
            except StopIteration:
                logging.error("Ran out of rhyme tags! Initialising new alphabet.")
                alphabet = iter(string.ascii_letters)
                rhyme_tag = next(alphabet)

        processed.append(
            dict(
                verse_id=idx,
                syllables=current_syllables,
                rhyme_score=rhyme_score,
                rhyme_tag=rhyme_tag,
            )
        )
    return processed


def collate_rhyme_scheme(annotated_stanza: list) -> str:
    """Join the rhyme tags rom each tagged verse to form a rhyme scheme."""
    return "".join(verse.get("rhyme_tag") for verse in annotated_stanza)


def get_stanzas_from_transcription(transcription: dict) -> list:
    """Parse a dict of transcribed verse lines and return a list of stanzas."""
    n_lines = len(transcription.keys()) - 1  # subtract the text_id key
    logging.debug("Number of lines in poem: %s", n_lines)
    poem = []
    stanza = []
    for n in range(n_lines):
        verse = transcription.get(f"line_{n}")
        if len(verse) > 0:
            syllables = utils.syllabify(verse)
            stanza.append(syllables)
        else:
            if len(stanza) == 0:
                continue
            poem.append(stanza)
            stanza = []
    if len(poem) == 0 and len(stanza) > 0:
        poem.append(stanza)
    return poem


def format_annotations(annotations: dict) -> dict:
    """Format lists of lists in the innermost dicts to a single line along with the dict key."""
    formatted_annotations = {}
    for stanza_key, stanza_value in annotations.items():
        formatted_stanza = []
        for verse in stanza_value:
            formatted_verse = {
                k: " ".join(map(str, v)) if isinstance(v, list) else v
                for k, v in verse.items()
            }
            formatted_stanza.append(formatted_verse)
        formatted_annotations[stanza_key] = formatted_stanza
    return formatted_annotations


def main(poem_file: str):
    """Annotate rhyming schemes in poems.

    Procedure:
    1. Split a poem into stanzas and verses.
    2. Transcribe the verses phonemically.
    3. Group the verse phonemes into syllables.
    4. Identify the last stressed syllable of a verse.
    5. Compare the syllable sequence of a verse with
        those of all previous verses in the same stanza. (NB! Replace "stanza" with "poem"? )
    6. Extract the rhyming part (i.e. nucleus + coda) of the last stressed syllable
    7. Score the rhyming:
        1="only rhymes match", # perfect match
        0.5="NØDRIM: onset also matches" # lame rhyme: 'tusenfryd' / 'fryd'
        0="No match"
    8. Tag the verse with a letter, depending on which line it rhymes with if at all
    """
    filepath = Path(poem_file)
    poem_text = json.loads(filepath.read_text())
    poem_id = poem_text.get("text_id")
    logging.debug("Tagging poem: %s", poem_id)
    stanzas = get_stanzas_from_transcription(poem_text)
    # print(stanzas)

    file_annotations = {}
    for idx, stanza in enumerate(stanzas):
        tagged = tag_rhyming_verses(stanza)
        rhyme_scheme = collate_rhyme_scheme(tagged)
        tagged.insert(0, {"rhyme_scheme": rhyme_scheme})
        file_annotations[f"stanza_{idx}"] = tagged

    formatted_content = format_annotations(file_annotations)
    outputfile = filepath.parent / f"{filepath.stem}_rhyme_scheme.json"
    with outputfile.open("w") as f:
        f.write(json.dumps(formatted_content, ensure_ascii=False, indent=4))

    logging.debug(
        "Saved rhyme scheme annotations for poem %s to \n\t%s", poem_id, outputfile
    )
    # Assume that the stanzas are independent of each other
    # and that the rhyme scheme is unique to each stanza


# %%

if __name__ == "__main__":
    import doctest
    doctest.testmod()

    import argparse
    from datetime import datetime

    parser = argparse.ArgumentParser(description="Tag rhyme schemes in a poem.")
    parser.add_argument(
        "jsonfile", type=str, help="Path to a json file with phonemic transcriptions."
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="Set logging level to debug."
    )
    args = parser.parse_args()

    if args.verbose:
        today = datetime.today().date()
        logging_file = f"{__file__.split('.')[0]}_{today}.log"
        logging.basicConfig(level=logging.DEBUG, filename=logging_file, filemode="a")
    main(args.jsonfile)



import pytest
from poetry_analysis.anaphora import detect_repeating_lines


def test_detect_repeating_lines_returns_both_linenumber_and_full_line():
    text = (
        "Idag er en fin dag\n" +
        "Hei på deg\n" +
        "Idag er en fin dag\n" +
        "bladibladibla\n" 
    )
    result = detect_repeating_lines(text)
    assert result == [
        ([0,2], "Idag er en fin dag"),
    ]



import pytest
from poetry_analysis.alliteration import extract_alliteration


def test_alliteration_returns_initial_consonant_counts():
    # Given
    text = [line.strip() for line in """Stjerneklare Septembernat
    Skaldene som siger sandhetden smukkest
    Ser Samla saga skriver.
    Litt flere ord
    """.splitlines()]

    expected = [
        {"line": 0, "symbol": "s", "count": 2, "words": ["Stjerneklare", "Septembernat"]},
        {"line": 1, "symbol": "s", "count": 5, "words": ["Skaldene", "som", "siger", "sandhetden", "smukkest"]},
        {"line": 2, "symbol": "s", "count": 4, "words": ["Ser", "Samla", "saga", "skriver."]},
        ]
    
    # When
    result = extract_alliteration(text)
    assert pytest.approx(expected) == result
  
def test_extract_alliteration_skips_punctuation(): 
    text = (
        "her har hun høstet hvaler\n"
        "eller er en eller annen linje\n"
        "... her.. var ... ellipse...\n"
        "- her - er - bindestrek-\n"
        "? hva med? spørsmålstegn ?\n"
    ).splitlines()

    result = extract_alliteration(text)
    assert len(result) ==  2
    assert result[0]["symbol"] == "h"
    assert result[1]["symbol"] == "e"
    assert not any(line["symbol"] in ["...", "-", "?"]  for line in result)



import pytest
from poetry_analysis.anaphora import extract_line_anaphora


@pytest.mark.parametrize(
    "text, expected_phrase, expected_count",
    [
        ("hello hello world", "hello", 2),
        ("hello world hello world hello world", "hello world", 3),
        ("Hei og hopp, hei og hallo, hei og hå", "hei og", 3),
        ("Hei hei hei hei hei", "hei", 5),
    ],
)
def test_extract_line_anaphora_extracts_longest_repeating_sequence(
    text, expected_phrase, expected_count
):
    result = extract_line_anaphora(text)
    actual = result[0]

    assert len(result) == 1
    assert actual["line_id"] == 0
    assert actual["phrase"] == expected_phrase
    assert actual["count"] == expected_count


def test_line_final_word_sequences_return_empty_list():
    text = (
        "Hei hello world hello world" + "\n"
        "Hey hello hello" + "\n\n"
        "hallo world world" + "\n"
        "Ai hello world"
    )
    result = extract_line_anaphora(text)
    assert result == []


def test_lines_without_anaphora_returns_empty_list():
    text = (
        "hello world" + "\n"
        "here we are" + "\n"
        "\n"
        "nothing to find here" + "\n"
        "only words that don't repeat"
    )
    result = extract_line_anaphora(text)
    assert result == []



import pytest
from poetry_analysis.anaphora import extract_poem_anaphora


@pytest.mark.skip("This behavious is not implemented yet")
def test_extract_poem_anaphora_merges_counts_across_stanzas():
    """Check that the longest repeating line initial word sequences are extracted.
    The count should be the number of lines the sequence is repeated in the stanza.
    """
    text = (
        "Hei på deg" + "\n"
        "Hei og hå" + "\n"
        "Hei sann" + "\n\n"
        "Hei hvor det går" + "\n"
        "Hei og hallo" + "\n"
    )
    expected = {
        "stanza_id": [0, 1],
        "line_id": [0, 1, 2, 3, 4],
        "phrase": "Hei",
        "count": 5,
    }

    result = extract_poem_anaphora(text)

    assert len(result) == 2

    assert all(r["stanza_id"] == e for r, e in zip(result, expected["stanza_id"]))
    actual_lineids = [r["line_id"] for r in result]
    assert all(
        actual_id in expected["line_id"] for ids in actual_lineids for actual_id in ids
    ), actual_lineids
    assert all(r["phrase"] == expected["phrase"] for r in result)
    assert sum(r["count"] for r in result) == expected["count"]


def test_poem_anaphora_returns_first_word_repeated_indices():
    text = (
        "jeg ser verden" + "\n"
        "jeg ser sola" + "\n"
        "jeg myser mot skyene" + "\n"
        "jeg ser havet" + "\n"
    )
    result = extract_poem_anaphora(text)

    assert len(result) == 1
    actual = result[0]
    assert actual["stanza_id"] == 0
    assert actual["line_id"] == [0, 1, 2, 3]
    assert actual["phrase"] == "jeg"
    assert actual["count"] == 4


def test_extract_stanza_anaphora_returns_empty_list_no_repeated_line_initial_phrase():
    text = (
        "hello world" + "\n"
        "here we are" + "\n"
        "\n"
        "nothing to find here" + "\n"
        "only words that don't repeat"
    )
    result = extract_poem_anaphora(text)
    assert result == []


def test_extract_multiple_anaphora_from_same_stanza():
    text = (
        "jeg ser på den hvite himmel,\n"
        "jeg ser på de gråblå skyer,\n"
        "jeg ser på denne blodige sol.\n"
        "dette er altså verden.\n"
        "dette er altså klodernes hjem.\n"
        "Dette er altså et kjent dikt.\n"
        "Her er vi i Norge.\n"
        "Her er vi på berget.\n"
        "Her er vi."
    )
    result = extract_poem_anaphora(text)
    assert len(result) == 3
    
    assert all(r["stanza_id"] == 0 for r in result)

    assert result[0]["phrase"] == "jeg"
    assert result[1]["phrase"] == "dette"
    assert result[2]["phrase"] == "her"
    assert all(r["count"] == 3 for r in result)



import re
import json
import string
from typing import Generator
from pathlib import Path

from convert_pa import nofabet_to_ipa, convert_nofabet


PUNCTUATION_MARKS = str(
    string.punctuation + "‒.,!€«»’”—⁷⁶⁰–‒––!”-?‒"
)  # Note! The three long dashes look identical, but are different unicode characters


def is_punctuation(char: str) -> bool:
    """Check if a character is a punctuation mark."""
    return char in PUNCTUATION_MARKS


def strip_redundant_whitespace(text: str) -> str:
    """Strip redundant whitespace and reduce it to a single space."""
    return re.sub(r"\s+", " ", text).strip()


def strip_punctuation(string: str) -> str:
    """Remove punctuation from a string"""
    alphanumstr = ""
    for char in string:
        if not is_punctuation(char):
            alphanumstr += char
    return strip_redundant_whitespace(alphanumstr)


def convert_to_syllables(phonemes: list, ipa=False) -> list:
    """Turn a sequence of phonemes into syllable groups."""
    transcription = phonemes if isinstance(phonemes, str) else " ".join(phonemes)
    if ipa:
        ipa = nofabet_to_ipa(transcription)
        syllables = ipa.split(".")
    else:
        syllables = convert_nofabet.nofabet_to_syllables(transcription)
    return syllables


def syllabify(transcription: list[list]) -> list:
    """Flatten list of syllables from a list of transcribed words."""
    syllables = [
        syll  # if syll is not None else "NONE"
        for word, pron in transcription
        for syll in convert_to_syllables(pron, ipa=False)
    ]
    return syllables


def annotate_transcriptions(transcription: list) -> Generator:
    for word, pronunciation in transcription:
        nofabet = format_transcription(pronunciation)
        yield dict(
            word=word,
            nofabet=nofabet,
            syllables=convert_nofabet.nofabet_to_syllables(nofabet),
            ipa=nofabet_to_ipa(nofabet),
        )


def split_paragraphs(text: str) -> list:
    """Split a text into paragraphs and paragraphs into lines."""
    return [
        [line.rstrip() for line in paragraph.rstrip().splitlines()]
        for paragraph in re.split("\n{2,}", text)
        if paragraph
    ]


def format_transcription(pronunciation):
    return " ".join(pronunciation)


def gather_stanza_annotations(func) -> callable:
    """Decorator to apply a function to each stanza in a text."""

    def wrapper(text: str) -> dict:
        stanzas = split_stanzas(text)
        stanza_annotations = {}
        for i, stanza in enumerate(stanzas, 1):
            stanza_text = "\n".join(stanza)
            stanza_annotations[f"stanza_{i}"] = func(stanza_text)
        return stanza_annotations

    return wrapper


def split_stanzas(text: str) -> list:
    """Split a poem into stanzas and stanzas into verses."""
    return [
        [verse.rstrip() for verse in stanza.rstrip().splitlines()]
        for stanza in re.split("\n{2,}", text)
        if stanza
    ]


def annotate(func, text: str, stanzaic: bool = False, outputfile: str | Path = None):
    if stanzaic:
        new_func = gather_stanza_annotations(func)
        annotations = new_func(text)
    else:
        annotations = func(text)
    if outputfile is not None:
        Path(outputfile).write_text(
            json.dumps(annotations, indent=4, ensure_ascii=False), encoding="utf-8"
        )
        print(f"Saved annotated data to {outputfile}")
    else:
        return annotations


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from poetry_analysis.anaphora import extract_stanza_anaphora


def test_extract_stanza_anaphora_returns_dict_with_single_element_list():
    text = [
        "hello world" + "\n",
        "here we are" + "\n\n",
        "nothing to find here" + "\n",
        "only words that don't repeat",
    ]
    result = extract_stanza_anaphora(text)
    assert len(result.keys()) == len(text)
    assert all(len(repetitions) == 1 for repetitions in result.values())


def test_extract_stanza_anaphora_returns_indices_of_all_occurrences(): 
    text = [
        "jeg ser på den hvite himmel,\n",
        "jeg ser på den hvite himmel,\n",
        "jeg ser på den hvite himmel,\n",
        "jeg ser på den hvite himmel,\n",
    ]
    result = extract_stanza_anaphora(text)
    assert len(result["jeg"]) == len(text)
    assert all(i == j for i, j in zip(result["jeg"], range(len(text))))


def test_extract_stanza_anaphora_skips_empty_line():
    text = ["",
        "du ser på den hvite himmel,\n",
        "",
        "jeg ser på den hvite himmel,\n",
        "jeg ser på den hvite himmel,\n",
    ]
    result = extract_stanza_anaphora(text)
    assert len(result["jeg"]) == 2
    assert result["jeg"] == [3, 4], result["jeg"]



import pytest
from poetry_analysis.rhyme_detection import score_rhyme


@pytest.mark.parametrize("syllable1, syllable2",[
    ([["F", "R", "YY1", "D"]], [["S", "YY1", "D"]]),
    ([["S", "II1", "N"]], [["D", "II2", "N"]]),
]) 
def test_score_rhyme_returns_1_when_rhyme_has_different_onsets(syllable1, syllable2): 
    result = score_rhyme(syllable1, syllable2)
    assert result == 1


@pytest.mark.parametrize("syllable1, syllable2",[
    ([["F", "R", "YY1", "D"]], [["F", "R", "YY1", "D"]]),
    ([["S", "II1", "N"]], [["S", "II1", "N"]]),
    ([["IH1"], ["G", "AH1", "NG"]], [["P", "OAH1"], ["G", "AH1", "NG"]])
]) 
def test_score_rhyme_returns_half_score_when_rhyme_has_same_onsets(syllable1, syllable2): 
    result = score_rhyme(syllable1, syllable2)
    assert result == 0.5


@pytest.mark.parametrize("syllable1, syllable2",[
    ([["F", "R", "YH1", "D"]], [["F", "R", "YY1", "D"]]),
    ([["S", "IH1", "N"]], [["S", "II0", "N"]]),
]) 
def test_score_rhyme_returns_0_when_rhyme_has_different_vowels(syllable1, syllable2): 
    result = score_rhyme(syllable1, syllable2)
    assert result == 0



"""Alliteration is the repetition of word-initial
consonants or consonant clusters.
"""

from pathlib import Path

from poetry_analysis.utils import annotate, gather_stanza_annotations


def count_alliteration(text: str) -> dict:
    """Count the number of times the same word-initial letter occurs in a text.

    Example use:
    >>> text = "Sirius som seer"
    >>> count_alliteration(text)
    {'s': 3}
    """
    words = text.split()
    initial_counts = {}

    for word in words:
        initial_letter = word[0].lower()
        if initial_letter in initial_counts:
            initial_counts[initial_letter] += 1
        else:
            initial_counts[initial_letter] = 1

    alliteration_count = {
        letter: count for letter, count in initial_counts.items() if count > 1
    }

    return alliteration_count


def extract_alliteration(text: list[str]) -> list[dict]:
    """Extract words that start with the same letter from a text.

    NB! This function is case-insensitive and compares e.g. S to s as the same letter.

    Args:
        text (list): A list of strings, where each string is a line of text.
    
    Example use:
    >>> text = ['Stjerneklare Septembernat Sees Sirius', 'Sydhimlens smukkeste Stjerne']
    >>> extract_alliteration(text)
    [{'line': 0, 'symbol': 's', 'count': 4, 'words': ['Stjerneklare', 'Septembernat', 'Sees', 'Sirius']}, {'line': 1, 'symbol': 's', 'count': 3, 'words': ['Sydhimlens', 'smukkeste', 'Stjerne']}]
    """

    alliterations = []

    for i, line in enumerate(text):   
        words = line.split() if isinstance(line, str) else line
        seen = {}
        for j, word in enumerate(words):
            initial_letter = word[0].lower()
            if not initial_letter.isalpha():
                continue

            if initial_letter in seen:
                seen[initial_letter].append(word)
            else:
                seen[initial_letter] = [word]

            if (j == len(words) - 1) and any(len(v) > 1 for v in seen.values()):
                alliteration_symbols = [k for k, v in seen.items() if len(v) > 1]
                for symbol in alliteration_symbols:
                    alliterations.append(
                        {
                            "line": i,
                            "symbol": symbol,
                            "count": len(seen[symbol]),
                            "words": seen[symbol],
                        }
                    )
                

    return alliterations


if __name__ == "__main__":
    # Test the functions with doctest
    import doctest
    doctest.testmod()

    # Parse user arguments
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("textfile", help="Filepath to the text to analyze.")
    parser.add_argument(
        "--split_stanzas", action="store_true", help="Split the text into stanzas."
    )
    parser.add_argument(
        "-o", "--outputfile", type=Path, help="File path to store results in. Defaults to the same file path and name as the input file, with the additional suffix `_alliteration.json`.",
    )
    args = parser.parse_args()

    # Analyze the text
    filepath = Path(args.textfile)
    text = filepath.read_text()

    if not args.outputfile:
        args.outputfile = Path(filepath.parent / f"{filepath.stem}_alliteration.json")
    annotate(extract_alliteration, text, stanzaic=args.split_stanzas, outputfile=args.outputfile)
